from reals._real import Real

from fractions import Fraction
from decimal import Decimal
from typing import Union


def sqrt(x: Union[Real, Fraction, int, Decimal]) -> Real:
    if not (isinstance(x, Real) or isinstance(x, Fraction) or isinstance(x, int) or isinstance(x, Decimal)):
        raise TypeError()
    return Real.from_number(x).__pow__(Fraction(1, 2))



from reals import Real, sqrt, e, pi
from reals.approximation import Approximation, best_rational_approximations

from reals._real import CachedComputation
from reals._algebraic_computation import AlgebraicComputation
from reals._quadratic_computation import QuadraticComputation

import pytest
from typing import Generator
from fractions import Fraction


def test_real_cache_is_shared() -> None:
    x = Real([1, 2, 3, 4, 5])

    c1 = x.compute()
    assert isinstance(c1, CachedComputation)

    c2 = x.compute()
    assert isinstance(c2, CachedComputation)

    next(c1)
    next(c1)
    next(c1)

    next(c2)

    assert x.cache == [1, 2, 3]
    assert c1.cache == [1, 2, 3]
    assert c2.cache == [1, 2, 3]

    assert c1.index == 3
    assert c2.index == 1


MAX_ITERATIONS = 100


def get_rational_from_real(x: Real) -> Fraction:
    a = Approximation(x)
    a.improve(MAX_ITERATIONS)
    fraction = a.lower_bound_fraction()
    assert fraction == a.upper_bound_fraction()
    assert fraction
    return fraction


def test_rational_real_terminates() -> None:
    x = Real.from_fraction(Fraction(123, 456))
    c = x.compute()
    terms = []
    for _ in range(0, 7):
        terms.append(next(c))

    assert terms == [0, 3, 1, 2, 2, 2, 2]

    with pytest.raises(StopIteration):
        next(c)


def test_terminating_algebraic_computation() -> None:
    x = Real.from_fraction(Fraction(1, 10))
    c = AlgebraicComputation(x.compute(), (2, 0, 0, 1))
    assert next(c) == 0
    assert next(c) == 5

    with pytest.raises(StopIteration):
        next(c)


def test_terminating_quadratic_computation() -> None:
    x = Real.from_fraction(Fraction(2, 1))
    y = Real.from_fraction(Fraction(1, 10))
    c = QuadraticComputation(x.compute(), y.compute(), (1, 0, 0, 0, 0, 0, 0, 1))
    assert next(c) == 0
    assert next(c) == 5

    with pytest.raises(StopIteration):
        next(c)

    with pytest.raises(StopIteration):
        next(c)


def test_terminating_quadratic_computation2() -> None:
    x = Real.from_fraction(Fraction(2, 1))
    y = Real.from_fraction(Fraction(1, 10))
    c = QuadraticComputation(x.compute(), y.compute(), (1, 0, 0, 0, 0, 0, 0, 1))
    assert next(c) == 0
    assert next(c) == 5

    with pytest.raises(StopIteration):
        next(c)


def test_termination() -> None:
    f = get_rational_from_real(Fraction(2, 1) * Real.from_fraction(Fraction(1, 10)))
    assert f == Fraction(1, 5)


def sqrt2_gen() -> Generator[int, None, None]:
    yield 1
    while True:
        yield 2


def test_generalized_continued_fractions() -> None:
    sqrt2 = Real(sqrt2_gen())
    a = Approximation(sqrt2 * sqrt2)
    a.improve_epsilon(Fraction(1, 1000))

    lo = a.lower_bound_fraction()
    assert lo

    hi = a.upper_bound_fraction()
    assert hi

    assert lo < 2 and 2 < hi


def test_sqrt_of_square_fractions() -> None:
    x = sqrt(Fraction(16, 9))
    assert x == Real.from_fraction(Fraction(4, 3))


def test_best_rational_approximations() -> None:
    best_rational_approximations(e / pi, 10)


def test_type_exception_when_mixing_with_floats() -> None:
    with pytest.raises(TypeError):
        _ = pi * 3.14



from reals._real import Real, Number
from reals._computation import Computation
from reals.approximation import Approximation
from reals._algebraic_computation import AlgebraicComputation

from decimal import Decimal
from fractions import Fraction
from typing import Generator


def log_frac_computation(x: int, y: int) -> Generator[tuple[int, int], None, None]:
    x -= y
    yield (0, x)

    m, n = x, y
    incr = 2 * y
    while True:
        yield (n, m)
        yield (2, m)
        m += x
        n += incr


def log_frac(f: Fraction) -> Real:
    x, y = f.as_integer_ratio()
    computation = Real(log_frac_computation(x, y)).compute()
    return Real(AlgebraicComputation(computation, (1, 0, 0, 1)))


class LogarithmicComputation(Computation):
    def __init__(self, x: Real):
        self.a = Approximation(x)
        self.n = 0
        self.increase_precision()

    def increase_precision(self):
        self.a.improve(10)
        self.lo = log_frac(self.a.lower_bound_fraction()).compute()
        self.hi = log_frac(self.a.upper_bound_fraction()).compute()
        for _ in range(0, self.n):
            next(self.lo)
            next(self.hi)

    def __next__(self) -> tuple[int, int]:
        while True:
            lower = next(self.lo)
            upper = next(self.hi)
            if lower == upper:
                self.n += 1
                return lower
            else:
                self.increase_precision()


def log(x: Number) -> Real:
    if isinstance(x, int) or isinstance(x, Fraction) or isinstance(x, Decimal):
        p, q = x.as_integer_ratio()
        return log_frac(Fraction(p, q))

    if isinstance(x, Real):
        return Real(LogarithmicComputation(x))

    raise TypeError()
